// comment of if you do not use/link to HDF5
#define ARMA_USE_HDF5 
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <iostream>
#include <boost/bind.hpp>
#include <boost/program_options.hpp>
#include "../mlstub/load.hpp"
#include "pcvmtrain.h"
#include "pcvmtrain_ny.h"
#include "kernel-matrix.h"
#include "nystroem-kernel-matrix.h"
#include "multiclassmodel_factory.h" 
#include "pcvmmulticlassmodel.h"
#include "pcvmmulticlassmodel_each.h"
//#include "pcvmmulticlassmodel_ecoc.h"*/
#include <assert.h>
#include <fstream>
#include <boost/serialization/serialization.hpp>
#include <boost/shared_ptr.hpp>
//#include "pcvm_serialization.hpp"

using namespace std;
using namespace arma;
namespace po   = boost::program_options;

template<typename T>
T mod(T a, int n)
{
    return a - floor(a/n)*n;
}   

int main(int argc, char** argv)
{
    // command line processing
	std::string sDataFilename = "./data/iris_n_3c.csv";
	std::string sModelFilename;
	std::string sCodingMatrixFilename;
	sModelFilename.clear();
	unsigned int kFold        = 5;
	bool bCrossvalidate       = false;
	bool bPredict	          = false;
	bool bNystroem	          = false;
	bool bHDF5	          = false;
	unsigned int iCycles	  = 500;
	int iKernel               = 0;
	int iMultiClassMode       = 0;
	arma::mat mCodingMatrix;
    try {
        po::options_description desc("Allowed options");
        desc.add_options()
            ("help", "produce help message")                                            
            ("data", po::value(&sDataFilename),"data filename (csv, last column labels)")            
			("k-fold", po::value(&kFold), "k for k-fold 0 (not set)")	    	    
		    ("cv", po::value(&bCrossvalidate), "crossvalidation (default false)")	    
			("cycles", po::value(&iCycles), "cycles (default 500)")
			("model", po::value(&sModelFilename), "model filename")
			("kernel", po::value(&iKernel), "kernel (0-linear (default), 1-elm-rbf, 2-input is a kernel")
			("nystroem", po::value(&bNystroem), "nystroem kernel (false)")
			("predict", po::value(&bPredict), "prediction (default false)")
			("hdf5", po::value(&bHDF5), "use hdf5 (binary) as input file format")
			("codematrix",  po::value(&sCodingMatrixFilename),"data filename for coding matrix - cvs file (for ecoc only)")
			("multiclassmode", po::value(&iMultiClassMode), "use 0 (one vs rest, default), 1 (each against each), 2 (ecoc)")
        ;

        po::variables_map vm;        
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);    

        if (vm.count("help")) {
            std::cout << desc << "\n";
            return 0;
        }    
	        
	std::cout << "Current settings ..." << std::endl;            		          				
	std::cout << "Data file: " << sDataFilename << std::endl;            					    
	std::cout << "Model file: " << sModelFilename << std::endl;
	if(iMultiClassMode ==2)
	{
			std::cout << "Coding matrix file: " << sCodingMatrixFilename << std::endl;
			mCodingMatrix.load(sCodingMatrixFilename);
	}		
	std::cout << "Cycles:  " << iCycles << std::endl;
	            					    	       	
	if(kFold > 1)
	{
	    std::cout << "k k-fold cv:  " << kFold << std::endl;	
	}
	std::cout << "Crossvalidation: " << (bCrossvalidate ? "yes" : "no") << std::endl;            					    		
	
    }
    catch(...) {
        std::cerr << "Exception of unknown type - process with defaults!\n";
  }
	
	

  // Load the dataset.
  arma::mat DataT;
  if(!bHDF5)
  {
  	mlpack_stub::data::Load(sDataFilename, DataT, true,false);
  }
  else
  {
	std::cout << "Load as hdf5 binary" << std::endl;
	DataT.load(sDataFilename, hdf5_binary);
	//dataset = dataset.t();
  }
  int M = DataT.n_rows;
  int N = DataT.n_cols;
    std::cout << "Have " << M << " rows and " << N << " columns" << std::endl;
	
  arma::vec Label = vectorise(DataT.col(N-1));  
  DataT = DataT.cols(0,N-2);
  pcvm_parameters oParameters;
  oParameters.SetMinimalThreshold(1e-5);
  oParameters.SetConvergenceThreshold(1e-3);
  oParameters.SetIterations (iCycles);

  IPCVMTrainInterface *oPCVMTrain = NULL;
  IKernelMatrixInterface *oKernelMatrix = NULL;
  if(bNystroem && iKernel == 2) // input is a Nystroem kernel)
  {
	oPCVMTrain    = new pcvm_train_ny();
	oKernelMatrix = new NystroemKernelMatrix();
  }
  else
  {
	oPCVMTrain    = new pcvm_train();
	oKernelMatrix = new KernelMatrix (iKernel);
  }	
  oPCVMTrain->SetParameters (oParameters );
  arma::vec vUniqueLabel = unique(Label);
	
  if(!bPredict)
  {
	  
	  bool bResult = false;
	  if(iKernel != 2)
	  {
			bResult = oKernelMatrix->Initialize(DataT,DataT);
	  }
	  else
	  {
			//mat RawKernelMatrix = DataT;
	        //RawKernelMatrix = RawKernelMatrix.cols(1,RawKernelMatrix.n_cols-1); 
			bResult = oKernelMatrix->Initialize(DataT,DataT,DataT);
	  }

	  assert(bResult);

	  MultiClassModelFactory oMultiClassModelFactory((multimodeltype)iMultiClassMode);
//	  IMultiClassModelInterface *const sFullModel = oMultiClassModelFactory.GetMultiClassModel();

	  boost::shared_ptr<IMultiClassModelInterface> sFullModel = oMultiClassModelFactory.GetMultiClassModel();// (new pcvm_multiclass_model());
	  if(iMultiClassMode == 2) 
	  {
		sFullModel->Init(mCodingMatrix);
	  }
	  sFullModel->GenerateMultiClassModels(*oPCVMTrain, *oKernelMatrix, DataT, Label);

	  // Save full model	 
	  if(!sModelFilename.empty())
	  {
		// Create an output archive

		std::ofstream ofs(sModelFilename.c_str());
		boost::archive::text_oarchive ar(ofs);
		// Write data
		ar & sFullModel;

		std::cout << "***" << std::endl;
		std::cout << sFullModel->GetUniqueLabels();
		std::cout << "***" << std::endl;
	  }	
	  
	  // predict the outcomes for the current full model
	  mat TrainingData = DataT.rows(sFullModel->GetReferenceIndices());	  
	  std::cout << "* Used model indices:" << sFullModel->GetReferenceIndices().t() << std::endl;
	  if(iKernel != 2)
		bResult = oKernelMatrix->Initialize(TrainingData,DataT);
	  else
	  {
		//mat RawKernelMatrix = DataT;		  
	    //RawKernelMatrix = RawKernelMatrix.cols(1,RawKernelMatrix.n_cols-1);	
		//std::cout << "Trainingdata refs agains all" << std::endl;
		//std::cout << TrainingData << std::endl;
		bResult = oKernelMatrix->Initialize(TrainingData,DataT,DataT); // DataT contains the (here) full input kernel matrix and hence also mC with landmark indication
		//std::cout << DataT.col(0) << std::endl;
		//std::cout << "Ny matrix:" << std::endl;
		//unsigned int NTemp, MTemp;
		//oKernelMatrix->size(NTemp,MTemp);
		//std::cout << oKernelMatrix->Evaluate(linspace<uvec>(0,NTemp-1,NTemp),linspace<uvec>(0,MTemp-1,MTemp)) << std::endl;
	  }
	  TrainingData.reset();
	  assert(bResult);
/*		std::cout << "Kernel on test:" << oKernelMatrix->GetKernelType() << std::endl;
	  std::cout << oKernelMatrix->GetKernelMatrix() << std::endl;
		std::cout << " *** " << std::endl;
*/	  mat PredictionResults = sFullModel->CalculateModelPredictions(*oKernelMatrix);  
	  uvec vPredictedModel  = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
	  
	  std::cout << (double)sum(abs(sFullModel->GetUniqueLabels()(vPredictedModel)-Label)<std::numeric_limits<double>::epsilon())/(double)Label.size() << std::endl;

	  if(bCrossvalidate)
	  {
			if(kFold>M)
			  kFold = M;
		  
			unsigned int iSetSize  = ceil(M/kFold);
			unsigned int iFullSize = kFold * iSetSize;
			uvec r = shuffle(linspace<uvec>(0,kFold-1,kFold));
			// V contains value between 0 and kFold-1 indicating whether an entry is included in the cv run or no
			uvec V = zeros<uvec>(M);
			uvec vShuffledIndices = shuffle(linspace<uvec>(0,M-1,M));
			if(M-iFullSize>0)
			{
				V(vShuffledIndices) =  join_vert(reshape(repmat(linspace<uvec>(0,kFold-1,kFold), iFullSize/kFold, 1), iFullSize, 1), r(linspace<uvec>(0,M-iFullSize-1,M-iFullSize)));
			}
			else
			{
				V(vShuffledIndices) =  reshape(repmat(linspace<uvec>(0,kFold-1,kFold), iFullSize/kFold, 1), iFullSize, 1);
			}

		  
			vec vPredictionAccuracy = zeros<vec>(kFold);

		  for(unsigned int k=0;k<kFold;++k)
			{
				// train model on training dat
				uvec vSelector = find(V!=k);
				
				if(iKernel != 2)
				{					
					oKernelMatrix->Initialize((DataT.rows(vSelector)),(DataT.rows(vSelector)));
				}
				else
				{						
					//mat RawKernelMatrix = DataT;
					//RawKernelMatrix = RawKernelMatrix.cols(1,RawKernelMatrix.n_cols-1); 
					bResult = oKernelMatrix->Initialize((DataT.rows(vSelector)),(DataT.rows(vSelector)),DataT);	
				}
				assert(bResult);
			
				mat DataSubT = (DataT.rows(vSelector));  
				vec SubLabel = Label(vSelector);

	  			MultiClassModelFactory oMultiClassModelFactoryCV((multimodeltype)iMultiClassMode);
				boost::shared_ptr<IMultiClassModelInterface> sFullModelCV = oMultiClassModelFactory.GetMultiClassModel();
	  			if(iMultiClassMode == 2) 
	  			{
					sFullModelCV->Init(mCodingMatrix);
	 	 		}

				sFullModelCV->GenerateMultiClassModels(*oPCVMTrain, *oKernelMatrix, DataSubT, SubLabel);
				DataSubT.reset();

				// make prediction on test data
				uvec vSelectorTest = find(V==k);

			    mat TrainingData = (DataT.rows(vSelector));
				TrainingData = TrainingData.rows(sFullModelCV->GetReferenceIndices());	

				if(iKernel != 2)
					bResult = oKernelMatrix->Initialize(TrainingData,(DataT.rows(vSelectorTest)));
				else
				{
					//mat RawKernelMatrix = DataT;
					//RawKernelMatrix = RawKernelMatrix.cols(1,RawKernelMatrix.n_cols-1); 
					bResult = oKernelMatrix->Initialize(TrainingData,(DataT.rows(vSelectorTest)),DataT);
				}
				
				assert(bResult);
				mat PredictionResults= sFullModelCV->CalculateModelPredictions(*oKernelMatrix);  			
			        uvec vPredictedModel = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
				vPredictionAccuracy(k) = (double)sum(abs(sFullModelCV->GetUniqueLabels()(vPredictedModel)-Label(vSelectorTest))<std::numeric_limits<double>::epsilon())/(double)vSelectorTest.size();
			}
			std::cout << mean(vPredictionAccuracy) << " " << sqrt(var(vPredictionAccuracy)) << std::endl; 
	  }
  }
  else if(!sModelFilename.empty())
  {
	  // Restore data
	  std::cout << "restore ..." << std::endl;
	  boost::shared_ptr<IMultiClassModelInterface> sFullModelRec;//( new pcvm_multiclass_model());
	  {
		// Create and input archive
		std::ifstream ifs(sModelFilename.c_str());
		boost::archive::text_iarchive ar(ifs);
		ar & sFullModelRec ;

/*		if(1)
{
		std::ofstream ofs("temp.model");
		boost::archive::text_oarchive ar2(ofs);
		// Write data
		dynamic_cast<pcvm_multiclass_model_each*>((sFullModelRec.get()))->sm_vvReferenceData.clear();
		ar2 & sFullModelRec;

}*/

		pcvm_model oModel1 = sFullModelRec->GetModels()[0]; // just choose the first model (we will not support different kernels on multiclass models but take always the same
	    	KernelMatrix oKernelMatrix(oModel1.GetKernelType());
		bool bResult = false;
		if(iKernel != 2)
		{
			const std::vector< std::vector<double> >& vvReferenceData = sFullModelRec->GetReferenceData();
			unsigned int NPoints = vvReferenceData.size();
			mat ReferenceData = zeros<mat>(NPoints,vvReferenceData[0].size());
			for(unsigned int k=0; k<NPoints; k++)
			{				
				arma::vec vTemp = arma::conv_to< arma::vec >::from(vvReferenceData[k]);
				ReferenceData.row(k)  = vTemp.t();
			}		  
			bResult = oKernelMatrix.Initialize(ReferenceData,DataT);
		}
		else
		{
			arma::uvec vReferenceIndices = sFullModelRec->GetReferenceIndices();
			unsigned int NPoints = vReferenceIndices.n_elem;
			/*mat RawKernelMatrix = DataT;
			RawKernelMatrix = RawKernelMatrix.cols(1,RawKernelMatrix.n_cols-1);
			RawKernelMatrix = RawKernelMatrix.cols(sFullModelRec.sm_vReferenceIndices);*/
			//std::cout << sFullModelRec->GetReferenceIndices() << std::endl;
			mat RawKernelMatrix = DataT.cols(vReferenceIndices+1); // +1 due to index column
			// first row as dummy - removed subsequently 
			RawKernelMatrix = join_vert(RawKernelMatrix.row(0),RawKernelMatrix);
			//std::cout << RawKernelMatrix << std::endl;
			bResult = oKernelMatrix.Initialize(conv_to<mat>::from(linspace<uvec>(0,NPoints-1,NPoints)),conv_to<mat>::from(linspace<uvec>(0,M-1,M)),RawKernelMatrix.t());
			/*std::cout << "Init km:" << std::endl;
			std::cout << oKernelMatrix.GetKernelMatrix()  <<std::endl;
			std::cout << "<-- Init km" << std::endl;*/
		}			 

/*

	  // predict the outcomes for the current full model
	  mat TrainingData = DataT.rows(sFullModel->GetReferenceIndices());	  
	  std::cout << "* Used model indices:" << sFullModel->GetReferenceIndices().t() << std::endl;
	  if(iKernel != 2)
		bResult = oKernelMatrix->Initialize(TrainingData,DataT);
	  else
	  {
		//mat RawKernelMatrix = DataT;		  
	    //RawKernelMatrix = RawKernelMatrix.cols(1,RawKernelMatrix.n_cols-1);		
		bResult = oKernelMatrix->Initialize(TrainingData,DataT,DataT);
	  }
	  TrainingData.reset();
	  assert(bResult);

	  uvec vPredictedModel  = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
	  
	  std::cout << (double)sum(abs(sFullModel->GetUniqueLabels()(vPredictedModel)-Label)<std::numeric_limits<double>::epsilon())/(double)Label.size() << std::endl;




*/

	  std::cout << "start prediction ..." << std::endl;
		assert(bResult);
		  
		mat PredictionResults    = sFullModelRec->CalculateModelPredictions(oKernelMatrix);
		vec vNormalizer          = sum(PredictionResults.cols(0,PredictionResults.n_cols-2),1);
		PredictionResults.cols(0,PredictionResults.n_cols-2)    	 = PredictionResults.cols(0,PredictionResults.n_cols-2) / repmat(vNormalizer,1,PredictionResults.n_cols-1);

		uvec vPredictedClassIndex    = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
        vec vPredictedClassLabel = sFullModelRec->GetUniqueLabels()(vPredictedClassIndex);		  
//		std::cout << PredictionResults << std::endl;
//		std::cout << sum(PredictionResults.t()) << std::endl;
		std::cout << vPredictedClassLabel << std::endl;
	  }
  }
  else
  {
		std::cout << "Need to specify a model file for prediction" << std::endl;
  }

  delete oPCVMTrain;
  delete oKernelMatrix;	
  return 0;
}
