#include <iostream>
#include <boost/bind.hpp>
#include <boost/program_options.hpp>
#include "../mlstub/load.hpp"
#include "pcvmtrain.h"
#include "pcvmtrain_ny.h"
#include "kernel-matrix.h"
#include "nystroem-kernel-matrix.h"
#include "pcvmmulticlassmodel.h"
#include <assert.h>
#include <fstream>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include "armaMex.hpp"
using namespace std;
using namespace arma;
namespace po   = boost::program_options;

template<typename T>
T mod(T a, int n)
{
    return a - floor(a/n)*n;
}   

void
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
  {
  // Check the number of input arguments.
  if (nrhs < 2)
    mexErrMsgTxt("Incorrect number of input arguments.");

  // Check type of input.
  if ( (mxGetClassID(prhs[0]) != mxDOUBLE_CLASS) || (mxGetClassID(prhs[1]) != mxDOUBLE_CLASS) )
  	  mexErrMsgTxt("First and second input must me of type double.");
  

  std::string sParameters = ""; // empty parameter string
  std::string sModelFilename;
  sModelFilename.clear();

  unsigned int kFold        = 5;
  bool bCrossvalidate       = false;
  bool bPredict	            = false;
  bool bNystroem	    = false;  
  unsigned int iCycles      = 500;
  int iKernel               = 2;

  if (nrhs == 2)
  {
  }
  else
  {
	if ( mxIsChar(prhs[2]) != 1) 
	{
		mexErrMsgTxt( "Third argument must be a parameter string");
	}
	char *input_buf;
	size_t buflen;
	buflen = (mxGetM(prhs[2]) * mxGetN(prhs[2])) + 1;

        /* copy the string data from prhs[0] into a C string input_ buf.    */
        input_buf = mxArrayToString(prhs[2]);
 
        if(input_buf == NULL) 
        	mexErrMsgTxt( "Could not convert input to string.");

 	sParameters = input_buf;
	std::istringstream buffer(sParameters);
	std::vector<std::string> vTemp((std::istream_iterator<std::string>(buffer)), 
                                 std::istream_iterator<std::string>());

	// initial parsing of the given parameters
	try {
        po::options_description desc("Allowed options");
        desc.add_options()
            ("help", "produce help message")                                            
			("k-fold", po::value(&kFold), "k for k-fold 0 (not set)")	    	    
		    ("cv", po::value(&bCrossvalidate), "crossvalidation (default false)")	    
			("cycles", po::value(&iCycles), "cycles (default 500)")
			("model", po::value(&sModelFilename), "model filename")
			("kernel", po::value(&iKernel), "kernel (0-linear (default), 1-elm-rbf, 2-input is a kernel")
			("nystroem", po::value(&bNystroem), "nystroem kernel (false)")
			("predict", po::value(&bPredict), "prediction (default false)")
        ;

	po::variables_map vm;
        po::store(po::command_line_parser(vTemp).options(desc).run(), vm);
        po::notify(vm);    

        if (vm.count("help")) {
            std::cout << desc << "\n";
            return;
        }    
	        
	std::cout << "Current settings ..." << std::endl;            		          				       					    
	std::cout << "Cycles:  " << iCycles << std::endl;
	            					    	       	
	if(kFold > 1)
	{
	    std::cout << "k k-fold cv:  " << kFold << std::endl;	
	}
	std::cout << "Crossvalidation: " << (bCrossvalidate ? "yes" : "no") << std::endl;            					    		
	
    	}
   	catch(...) {
        	std::cerr << "Exception of unknown type - process with defaults!\n";
	}
	


	mxFree(input_buf);
  }

  // Check if input is real.
  if ( (mxIsComplex(prhs[0])) || (mxIsComplex(prhs[1])) )
    mexErrMsgTxt("Input must be real.");
  
  // Create matrices X and Y from the first and second argument.
  mat  Data  = armaGetPr(prhs[0]);
  vec  Label = conv_to<vec>::from(armaGetPr(prhs[1]));
  Data       = Data.t();
  unsigned int M = Data.n_rows;
  unsigned int N = Data.n_cols;
  
  pcvm_parameters oParameters;
  oParameters.SetMinimalThreshold(1e-3);
  oParameters.SetConvergenceThreshold(1e-3);
  oParameters.SetIterations (iCycles);

  IPCVMTrainInterface *oPCVMTrain = NULL;
  IKernelMatrixInterface *oKernelMatrix = NULL;
  if(bNystroem && iKernel == 2) // input is a Nystroem kernel)
  {
	oPCVMTrain    = new pcvm_train_ny();
	oKernelMatrix = new NystroemKernelMatrix();
  }
  else
  {
	oPCVMTrain    = new pcvm_train();
	oKernelMatrix = new KernelMatrix (iKernel);
  }	
  oPCVMTrain->SetParameters (oParameters );
  arma::vec vUniqueLabel = unique(Label);
	
  if(!bPredict)
  {
	  
	  bool bResult = false;
	  if(iKernel != 2)
		bResult = oKernelMatrix->Initialize(Data,Data);
	  else
	  {
		bResult = oKernelMatrix->Initialize(Data,Data,Data);
	  }
	  assert(bResult);

	  // run directly the multiclass 1 vs rest wrapper
	  pcvm_multiclass_model sFullModel;
	  sFullModel.GenerateMultiClassModels(*oPCVMTrain, *oKernelMatrix, Data, Label);
	  
	  // Save full model	 
	  if(!sModelFilename.empty())
	  {
		// Create an output archive
		std::ofstream ofs(sModelFilename.c_str());
		boost::archive::text_oarchive ar(ofs);

		// Write data
		ar & sFullModel;
	  }	
	  
	  // predict the outcomes for the current full model
	  mat TrainingData = Data.rows(sFullModel.sm_vReferenceIndices);
	  std::cout << "* Used model indices:" << sFullModel.sm_vReferenceIndices.t() << std::endl;
	  if(iKernel != 2)
		bResult = oKernelMatrix->Initialize(TrainingData,Data);
	  else
	  {
		bResult = oKernelMatrix->Initialize(TrainingData,Data,Data);
	  }
 	  TrainingData.reset();
	  assert(bResult);
	  
	  mat PredictionResults = sFullModel.CalculateModelPredictions(*oKernelMatrix);  
	  uvec vPredictedModel  = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
	  
	  std::cout << (double)sum(abs(sFullModel.sm_vUniqueLabel(vPredictedModel)-Label)<std::numeric_limits<double>::epsilon())/(double)Label.size() << std::endl;

	  if(bCrossvalidate)
	  {
			if(kFold>M)
			  kFold = M;
		  
			unsigned int iSetSize  = ceil(M/kFold);
			unsigned int iFullSize = kFold * iSetSize;
			uvec r = shuffle(linspace<uvec>(0,kFold-1,kFold));
			// V contains value between 0 and kFold-1 indicating whether an entry is included in the cv run or no
			uvec V = zeros<uvec>(M);
			uvec vShuffledIndices = shuffle(linspace<uvec>(0,M-1,M));
			if(M-iFullSize>0)
			{
				V(vShuffledIndices) =  join_vert(reshape(repmat(linspace<uvec>(0,kFold-1,kFold), iFullSize/kFold, 1), iFullSize, 1), r(linspace<uvec>(0,M-iFullSize-1,M-iFullSize)));
			}
			else
			{
				V(vShuffledIndices) =  reshape(repmat(linspace<uvec>(0,kFold-1,kFold), iFullSize/kFold, 1), iFullSize, 1);
			}

			vec vPredictionAccuracy = zeros<vec>(kFold);
			for(unsigned int k=0;k<kFold;++k)
			{
				// train model on training dat
				uvec vSelector = find(V!=k);
				
				if(iKernel != 2)
					oKernelMatrix->Initialize((Data.rows(vSelector)),(Data.rows(vSelector)));
				else
				{
					bResult = oKernelMatrix->Initialize((Data.rows(vSelector)),(Data.rows(vSelector)),Data);	
				}
				assert(bResult);
				
				pcvm_multiclass_model sFullModelCV;
				sFullModelCV.GenerateMultiClassModels(*oPCVMTrain, *oKernelMatrix, (Data.cols(vSelector)).t(), Label(vSelector));

				// make prediction on test data
				uvec vSelectorTest = find(V==k);

				mat TrainingData = (Data.rows(vSelector));
				TrainingData = TrainingData.rows(sFullModelCV.sm_vReferenceIndices);	

				if(iKernel != 2)
					bResult = oKernelMatrix->Initialize(TrainingData,(Data.rows(vSelectorTest)));
				else
				{
					bResult = oKernelMatrix->Initialize(TrainingData,(Data.rows(vSelectorTest)),Data);
				}
				
				assert(bResult);
				mat PredictionResults= sFullModelCV.CalculateModelPredictions(*oKernelMatrix);  			
			        uvec vPredictedModel = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
				vPredictionAccuracy(k) = (double)sum(abs(sFullModelCV.sm_vUniqueLabel(vPredictedModel)-Label(vSelectorTest))<std::numeric_limits<double>::epsilon())/(double)vSelectorTest.size();
			}
			std::cout << mean(vPredictionAccuracy) << " " << sqrt(var(vPredictionAccuracy)) << std::endl; 

  			// Create the output argument plhs[0] to return cube C
  			plhs[0] = armaCreateMxMatrix(1, 2);
  
	  		// Return the cube C as plhs[0] in Matlab/Octave
			mat CVSummary  = zeros(1,2);
			CVSummary(0,0) = mean(vPredictionAccuracy);
			CVSummary(0,1) = sqrt(var(vPredictionAccuracy));
			armaSetPr(plhs[0], CVSummary);

	  }
	  else
	  {
  		// Create the output argument plhs[0] to return cube C
  		plhs[0] = armaCreateMxMatrix(PredictionResults.n_rows, PredictionResults.n_cols);
  
  		// Return the cube C as plhs[0] in Matlab/Octave
		armaSetPr(plhs[0], PredictionResults);
	  }

  }
  else if(!sModelFilename.empty())
  {
	  // Restore data
	  pcvm_multiclass_model sFullModelRec;
	  {
		// Create and input archive
		std::ifstream ifs(sModelFilename.c_str());
		boost::archive::text_iarchive ar(ifs);
		// Load
		ar & sFullModelRec;

		pcvm_model oModel1 = sFullModelRec.sm_Model[0]; // just choose the first model (we will not support different kernels on multiclass models but take always the same
		std::vector< std::vector<double> > vvReferenceData = sFullModelRec.sm_vvReferenceData;
		unsigned int NPoints = vvReferenceData.size();
		mat ReferenceData = zeros<mat>(NPoints,vvReferenceData[0].size());
		for(unsigned int k=0; k<NPoints; k++)
		{				
			arma::vec vTemp = arma::conv_to< arma::vec >::from(vvReferenceData[k]);
			ReferenceData.row(k)  = vTemp.t();
		}		  
	    KernelMatrix oKernelMatrix(oModel1.GetKernelType());
		bool bResult = false;
		if(iKernel != 2)
			bResult = oKernelMatrix.Initialize(ReferenceData,Data);
		else
		{
			mat RawKernelMatrix = Data.cols(sFullModelRec.sm_vReferenceIndices+1);
			// first row as dummy - removed subsequently 
			RawKernelMatrix = join_vert(RawKernelMatrix.row(0),RawKernelMatrix);
			bResult = oKernelMatrix.Initialize(conv_to<mat>::from(linspace<uvec>(0,NPoints-1,NPoints)),conv_to<mat>::from(linspace<uvec>(0,M-1,M)),RawKernelMatrix.t());
		}			 
		assert(bResult);
		  
		mat PredictionResults    = sFullModelRec.CalculateModelPredictions(oKernelMatrix);
		uvec vPredictedModel     = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
    		vec vPredictedClassLabel = sFullModelRec.sm_vUniqueLabel(vPredictedModel);
		std::cout << vPredictedClassLabel << std::endl;
  
  		// Create the output argument plhs[0] to return cube C
  		plhs[0] = armaCreateMxMatrix(PredictionResults.n_rows, PredictionResults.n_cols);
  
  		// Return the cube C as plhs[0] in Matlab/Octave
		armaSetPr(plhs[0], PredictionResults);

	  }
  }
  else
  {
		std::cout << "Need to specify a model file for prediction" << std::endl;
  }

  delete oPCVMTrain;
  delete oKernelMatrix;	  
  return;
  }
