/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * pcvmmulticlassmodel.cc
 * Copyright (C) 2015 Frank-Michael Schleif <schleify@cs.bham.ac.uk>
 *
 * pcvm_cpp is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pcvm_cpp is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "pcvmmulticlassmodel.h"
#include "math_stat_helper.hpp"
#include <vector>
#include <boost/bind.hpp>
#include <boost/serialization/export.hpp>
//BOOST_CLASS_EXPORT(pcvm_multiclass_model)

using namespace std;

mat pcvm_multiclass_model::CalculateModelPredictions(const IKernelMatrixInterface &oKernelInstance) const
{
  arma::vec vUniqueLabelFullModel     = sm_vUniqueLabel;
  std::vector<pcvm_model> vPCVMModels = sm_Model;
  std::vector<vec> vPredictions;
  arma::uvec ReferenceIndices         = sm_vReferenceIndices;
  unsigned int iRows = 0;
  unsigned int iCols = 0;
  oKernelInstance.size(iRows,iCols);
  unsigned int N = iCols;
  vPredictions.resize(vUniqueLabelFullModel.size());
  
  transform(vPCVMModels.begin(),vPCVMModels.end(),vPredictions.begin(),boost::bind(&pcvm_model::CalculatePrediction,_1,boost::cref(oKernelInstance),ReferenceIndices));
  mat mPredictions = zeros<mat>(N,vUniqueLabelFullModel.size());
  for(std::vector<vec>::iterator iter=vPredictions.begin();
  	  iter != vPredictions.end(); ++iter)
		mPredictions.col(std::distance(vPredictions.begin(),iter)) = *iter;
  //std::cout << mPredictions << std::endl;
  mat mProbabilities  = normalCDF(-mPredictions);
  //std::cout << mProbabilities << std::endl;
  uvec vPredictedModel = zeros<uvec>(N);  
  for(int i=0; i!=N; ++i)
  {
	  mat vRow = mProbabilities.row(i);
	  vPredictedModel(i) = std::distance(vRow.begin(),max_element(vRow.begin(),vRow.end()));
  }
  mat PredictionResults = zeros<mat>(N,vUniqueLabelFullModel.size()+1);
  PredictionResults.cols(0,vUniqueLabelFullModel.size()-1) = mProbabilities;
  PredictionResults.col(vUniqueLabelFullModel.size())     = conv_to< vec >::from(vPredictedModel);
  return PredictionResults;
}

pcvm_model pcvm_multiclass_model::CalculateClassificationModel (int i,IPCVMTrainInterface &oPCVMTrain,const IKernelMatrixInterface &oKernel, const vec &Label) const  
{  
	vec vNewLabelVector  = ones<vec>(Label.size(),1);
	uvec vValid = find(Label == i);
	vNewLabelVector(vValid) = -vNewLabelVector(vValid);
	oPCVMTrain.Apply(oKernel,vNewLabelVector);
	std::cout << "*";
	std::cout << std::flush;
	return oPCVMTrain.GetModel();
}


void pcvm_multiclass_model::GenerateMultiClassModels(IPCVMTrainInterface &oPCVMTrain,const IKernelMatrixInterface &oKernel, const mat &Data, const vec &Label)
{
	pcvm_multiclass_model &sFullModel = *this;
	arma::vec vUniqueLabel = unique(Label);
	std::vector<pcvm_model> vPCVMModels;
	vPCVMModels.resize(vUniqueLabel.size());
	std::cout << " - " << std::endl;
	transform(vUniqueLabel.begin(),vUniqueLabel.end(),vPCVMModels.begin(),boost::bind(&pcvm_multiclass_model::CalculateClassificationModel,this,_1,boost::ref(oPCVMTrain),boost::cref(oKernel),Label));
	std::cout << " - " << std::endl;
	sFullModel.sm_vUniqueLabel = vUniqueLabel;
	sFullModel.sm_Model        = vPCVMModels;
	int iKernel = oKernel.GetKernelType();
	// identify the reference data points we have to keep by
	// analyzing the support indices of the various models
	arma::uvec vUniqueSupportIndices; 
	for(unsigned int k=0; k<vPCVMModels.size(); ++k)
	{
		if(k==0)
			vUniqueSupportIndices =	vPCVMModels[k].GetSupportIndices ();
		else
			vUniqueSupportIndices = unique(join_vert(vUniqueSupportIndices,vPCVMModels[k].GetSupportIndices ()));
	}
	sFullModel.sm_vReferenceIndices = vUniqueSupportIndices; 
				
	// reference data are only needed for vectorial data
	// so we keep sFullModel.sm_vvReferenceData
	if(iKernel != 2)
	{					
		int NPoints = vUniqueSupportIndices.n_elem;
		typedef std::vector<double> stdvec;
		std::vector< stdvec > vvData;
		vvData.resize(NPoints);
		for(int k=0; k<NPoints; k++)
		{
			arma::vec vTemp    = vectorise(Data.row(vUniqueSupportIndices(k)));
			vvData[k] = arma::conv_to< stdvec >::from(vTemp);
		}	
		sFullModel.sm_vvReferenceData = vvData;
	}
	else
	{
		sFullModel.sm_vvReferenceData.clear();
	}
}

/*
template < class Archive>
		void pcvm_multiclass_model::save(Archive& ar, const unsigned version) const {
			const pcvm_multiclass_model& oClass = *this;
			typedef std::vector<double> stdvec;
			typedef std::vector<unsigned int> uintvec;
			stdvec vDoubleVec;		
			uintvec vVec2;
			vDoubleVec = conv_to< stdvec  >::from(oClass.sm_vUniqueLabel);
			vVec2      = conv_to< uintvec >::from(oClass.sm_vReferenceIndices);
			//std::copy(vDoubleVec.begin(),vDoubleVec.end(),std::ostream_iterator<double>(std::cout," [multi-s] "));
			ar & vDoubleVec;  // Simply serialize the data members of Obj
			ar & vVec2;
			ar & oClass.sm_Model;
			ar & oClass.sm_vvReferenceData;
}
*/

/*template < class Archive>
			void pcvm_multiclass_model::load(Archive& ar, unsigned int) {
			const pcvm_multiclass_model& oClass = *this;
			std::vector<double> vDoubleVec;
			std::vector<unsigned int> vVec2;
			ar & vDoubleVec;
			ar & vVec2;
			//std::copy(vDoubleVec.begin(),vDoubleVec.end(),std::ostream_iterator<double>(std::cout," [multi-l] "));
			oClass.sm_vUniqueLabel = zeros<vec>(vDoubleVec.size());
			oClass.sm_vUniqueLabel = conv_to< arma::vec >::from(vDoubleVec);
			oClass.sm_vReferenceIndices = conv_to< arma::uvec >::from(vVec2);
			ar & oClass.sm_Model;
			ar & oClass.sm_vvReferenceData;
}
*/

