/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * pcvmtrain.cc
 * Copyright (C) 2015 Frank-Michael Schleif <schleify@cs.bham.ac.uk>
 *
 * This is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "pcvmtrain.h"
#include "math_stat_helper.hpp"
#include <cmath>

using namespace arma;
pcvm_train::pcvm_train()
{
	m_oParameters = pcvm_parameters ();
	m_oPCVMModel  = pcvm_model (zeros<vec>(1,1),0.0); // dummy model
}

pcvm_train::~pcvm_train()
{
}

void pcvm_train::SetParameters(pcvm_parameters oParameters)
{
	m_oParameters = oParameters;
}

pcvm_model pcvm_train::GetModel(void)
{
	return m_oPCVMModel;	
}

void pcvm_train::Apply(const IKernelMatrixInterface &oKernel,const vec &Label)
{
	// the following is a c++ implementation of the original Matlab code of Huanhuan Chen
	// of the Probabilistic Classification Vector Machine
	// Huanhuan Chen, Peter Tino and Xin Yao. Probabilistic Classification Vector Machines. 
	// IEEE Transactions on Neural Networks. vol.20, no.6, pp.901-914, June 2009. 
	// Copyright (c) Huanhuan Chen and Frank-Michael Schleif

	// the maximal iterations
	const mat &Data = oKernel.GetKernelMatrix ();
	unsigned int niters = this->m_oParameters.GetIterations ();
	unsigned int ndata  = Data.n_rows;

	vec w;
	vec Ht;
	double b;
	if(1)
	{
		// Initial weight vector to let w to be large than zero
		w = randu(ndata,1) + 0.2;

		// Initial bias b
		b = as_scalar(randu(1,1));


		// initialize the auxiliary variables Ht to follow the target labels of the training set
		Ht = 10*randu(ndata,1) % Label + randu(ndata,1); // % is elementwise multiplication in armadillo ...
	}
	else
	{	
		// some test code
		w <<  0.5 << 1 << 1 << 0.9 << 0.3;
		b = 0.85;
		Ht = 10*w % Label + w;
	}
	
	// Threshold to determine whether this is small
	double w_minimal = this->m_oParameters.GetMinimalThreshold ();

	// Threshold for convergence 
	double threshold = this->m_oParameters.GetConvergenceThreshold ();

	// all one vector
	vec I = ones<vec>(ndata);

	vec y = Label;

	// active vector indicator
	uvec nonZero = linspace<uvec>(0, ndata-1, ndata);

	// non-zero weight vector
	vec w_nz = w;

	vec wold = w;

	mat K = Data;

	// scale columns of kernel matrix with label trainY - this is better than
	// scaling the weights since the weights are changed and truncated so we
	// loose control on the y's otherwise
	mat Ky = K % repmat(Label.t(), ndata, 1);

	// Main loop of algorithm
	for(int n = 1; n<=niters; n++)
	{
		    // non-zero vector
			mat Ky_nz = Ky.cols((nonZero));
			mat Ht_nz;
			if(n==1)
    			Ht_nz = Ht;
			else
    			Ht_nz = Ky_nz*w_nz + b * ones<vec>(ndata); // hidden variables provided by the linear model

		    // -- Expectation step --
			// expectation for H   
		    mat Z = Ht_nz + y % normalPDF(Ht_nz)/(normalCDF(y % Ht_nz)+ std::numeric_limits<double>::epsilon()); // see Eq 45, Z = expectation on H  equivalent to Z = z psi(y*z) + z * pdf(z)
			
			// expectation for b
			double S = sqrt(2)*abs(b);        // --- beta parameter in a gaussian prior, see Eq. 47

			// expectation for a Adjust the new estimates for the parameters
			mat M = sqrt(2)*diagmat(w_nz);   // sqrt(1./alpha) = (1./(0.5 / w^2))^0.5 = sqrt(2) * w --- alpha parameter in a gaussian prior, see Eq 46, see also text before Eq 51
			
			// -- Maximization step --
			// new weight vector
			mat Hess = eye(M.n_rows,M.n_rows)+M * (Ky_nz.t()*Ky_nz) * M; // % see Eq 51 (M * phi^t * phi * M + I)
			mat U    = chol(Hess);                                    // % if (chol(A)) is psd than U'*U = A      (is always psd due to K'*K in Hess)
			mat Ui   = inv(U);

			
			w(nonZero) = M*(Ui*Ui.t())*M*(Ky_nz.t()*Z - b*Ky_nz.t()*I);     // % Eq 51 , A^-1 = (U'*U)^-1 = Ui * Ui'
			b = S*pow(1+ S*ndata*S,-1)*S*as_scalar(I.t()*Z - I.t()*Ky_nz*w((nonZero))); // % Eq 52                             

			nonZero	      = find(w > w_minimal);

			// determine used vectors
			uvec used     = nonZero;

			vec wNew      = zeros(ndata,1);
			wNew(nonZero) = w(nonZero);
			w             = wNew;
		
			// non-zero weight vector
			w_nz          = w(nonZero);	

			if (n >1 && max(abs(w - wold))< threshold)
    			break;
			else
    			wold = w; 
		
	}
	m_oPCVMModel = pcvm_model(w % Label,b,oKernel.GetKernelType()); 
}

pcvm_parameters pcvm_train::GetParameters(void)
{
	return m_oParameters;
}
