/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * pcvmtrain_ny.cc
 * Copyright (C) 2015 Frank-Michael Schleif <schleify@cs.bham.ac.uk>
 *
 * This is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "pcvmtrain_ny.h"
#include "math_stat_helper.hpp"
#include "pinv_ny_sym.hpp"
#include "nystroem-kernel-matrix.h"
#include <cmath>

using namespace arma;
pcvm_train_ny::pcvm_train_ny()
{
	m_oParameters = pcvm_parameters ();
	m_oPCVMModel  = pcvm_model (zeros<vec>(1,1),0.0); // dummy model
}

pcvm_train_ny::~pcvm_train_ny()
{
}

void pcvm_train_ny::SetParameters(pcvm_parameters oParameters)
{
	m_oParameters = oParameters;
}

pcvm_model pcvm_train_ny::GetModel(void)
{
	return m_oPCVMModel;	
}

void pcvm_train_ny::Apply(const IKernelMatrixInterface &oKernel,const vec &Label)
{
	// the following is a c++ implementation of the original Matlab code of Huanhuan Chen
	// of the Probabilistic Classification Vector Machine
	// Huanhuan Chen, Peter Tino and Xin Yao. Probabilistic Classification Vector Machines. 
	// IEEE Transactions on Neural Networks. vol.20, no.6, pp.901-914, June 2009. 
	// Copyright (c) Huanhuan Chen and Frank-Michael Schleif

	// the maximal iterations
	//const NystroemKernelMatrix *oNystroemKernelMatrix = dynamic_cast<const NystroemKernelMatrix*>(&oKernel);
	unsigned int niters = this->m_oParameters.GetIterations ();
	unsigned int rows, cols;
	oKernel.size(rows,cols);
	unsigned int ndata  = rows;
	
	vec w;
	vec Ht;
	double b;

	// Initial weight vector to let w to be large than zero
	w = randu(ndata,1) + 0.2;
	
	
	// little addon to account for imbalanced rest classes
/*if(1)
{
	uvec vTargetClass = find(Label<0);
	unsigned int N1   = vTargetClass.n_elem;
	uvec vRestClass   = find(Label>0);
	unsigned int NR   = vRestClass.n_elem;
	vec vIndicator    = randu(NR,1); 
	double dRatio     = 2*(double)N1/(double)NR; 
	if(dRatio<1.0) // rest class is substanially larger than target class ?
	{   
		vRestClass         = vRestClass(find(vIndicator>=dRatio));
		w.rows(vRestClass) = zeros<vec>(vRestClass.n_elem); // remove a large random fraction of rest points such that not more than 2*N1 are kept
	}
}*/
	
	// Initial bias b
	b = as_scalar(randu(1,1));


	// initialize the auxiliary variables Ht to follow the target labels of the training set
	Ht = 10*randu(ndata,1) % Label + randu(ndata,1); // % is elementwise multiplication in armadillo ...

	
	// Threshold to determine whether this is small
	double w_minimal = this->m_oParameters.GetMinimalThreshold ();

	// Threshold for convergence 
	double threshold = this->m_oParameters.GetConvergenceThreshold ();

	// all one vector
	vec I = ones<vec>(ndata);

	vec y = Label;

	// active vector indicator
	// uvec nonZero = linspace<uvec>(0, ndata-1, ndata);
	uvec nonZero = find(w > w_minimal);

	// non-zero weight vector
	// vec w_nz = w;
	vec w_nz = w(nonZero);	

	vec wold = w;

	mat Wpinv = oKernel.GetParameter (2);

	const mat K1 = oKernel.GetParameter (1); // this is a costly copy ...
	const mat K2 = Wpinv * K1.t(); // this is likely to be costly due to the transpose ...
	Wpinv.reset(); 
	
	const mat &K1FullIndex = K1; 
	const mat &K2FullIndex = K2; 

	// Main loop of algorithm
	for(int n = 1; n<=niters; n++)
	{
			int NrNonZero     = nonZero.n_elem;
			vec CurrentTrainY = y.rows(nonZero);
		    
			mat Ht_nz;
			if(n==1)
    			Ht_nz = Ht;
			else
			{
		        vec vTemp      = (CurrentTrainY % w_nz);
    			mat Temp       = vTemp.t() * K1.rows(nonZero); 
    			mat Ht_nzPart2 = (Temp * K2FullIndex).t();
				Temp.reset();
    			Ht_nz          = Ht_nzPart2 + b* ones<vec>(ndata); // hidden variables provided by the linear model
			}


			// -- Expectation step --
			// expectation for H   
			mat Z = Ht_nz + y % normalPDF(Ht_nz)/(normalCDF(y % Ht_nz)+ std::numeric_limits<double>::epsilon()); // see Eq 45, Z = expectation on H  equivalent to Z = z psi(y*z) + z * pdf(z)
			
			// expectation for b
			double S = sqrt(2)*abs(b);        // --- beta parameter in a gaussian prior, see Eq. 47

		
		    if(NrNonZero<100)
			{
    			mat Knm = K1FullIndex*K2.cols(nonZero); 
					
				// expectation for a Adjust the new estimates for the parameters
				mat M = sqrt(2)*diagmat(w_nz);   // sqrt(1./alpha) = (1./(0.5 / w^2))^0.5 = sqrt(2) * w --- alpha parameter in a gaussian prior, see Eq 46, see also text before Eq 51
			
				// -- Maximization step --
				// new weight vector
				mat TempY = (repmat(CurrentTrainY.t(),NrNonZero,1) % repmat(CurrentTrainY,1,NrNonZero));
    			// Hess = eye(size(M,1))+M*( (Knm'*Knm)  .* bsxfun(@times,CurrentTrainY,CurrentTrainY') )*M;                  % see Eq 51 (M * phi^t * phi * M + I)    				
				mat Hess = eye(M.n_rows,M.n_rows)+M * ((Knm.t()*Knm) % TempY) * M; // % see Eq 51 (M * phi^t * phi * M + I)
				TempY.reset();
				mat U    = chol(Hess);                                    // % if (chol(A)) is psd than U'*U = A      (is always psd due to K'*K in Hess)				
				Hess.reset();
				mat Ui   = inv(U);
				U.reset();
			    //        w(IndexNonZeroLogical) = M*(Ui*Ui')*M*(CurrentTrainY .* (Knm'*Z)- b*(CurrentTrainY .* (Knm'*I)));    
				w(nonZero) = M*(Ui*Ui.t())*M*(CurrentTrainY % (Knm.t()*Z) - b*(CurrentTrainY % (Knm.t()*I)));     // % Eq 51 , A^-1 = (U'*U)^-1 = Ui * Ui'
			}
			else
			{
    			// expectation for a Adjust the new estimates for the parameters
    			vec vT               = sqrt(2)*(w_nz);  

				uvec r               = shuffle(linspace<uvec>(0,NrNonZero-1,NrNonZero));
				uvec vLandmarksInner = r.rows(0,min(NrNonZero-1,500)); // 50 landmarks 
		        mat  Temp            = K1FullIndex*K2.cols(nonZero(vLandmarksInner));
    			Temp	             = (K2FullIndex * Temp);
				mat  Cnm             = K1.rows(nonZero) * Temp;

				Temp.reset();
				
				//         Cnm = Cnm .* bsxfun(@times,CurrentTrainY,CurrentTrainY(R)');
				mat  Temp3           = repmat(CurrentTrainY(vLandmarksInner).t(),NrNonZero,1) % repmat(CurrentTrainY,1,vLandmarksInner.n_elem);
    			Cnm                  = Cnm % Temp3;

				mat Eye              = zeros(NrNonZero,vLandmarksInner.n_elem);
				Eye.rows(vLandmarksInner) = eye(vLandmarksInner.n_elem,vLandmarksInner.n_elem);
				//         Cnm   = Eye + bsxfun(@times,(bsxfun(@times,Cnm,vT)) ,vT(R)');
				Temp3                = repmat(vT,1,vLandmarksInner.n_elem) % Cnm;				

				Cnm                  = Eye + Temp3 % repmat(vT(vLandmarksInner).t(),NrNonZero,1);

				Temp3.reset();
				Eye.reset();
				
				mat Wpinv			 = pinv(Cnm.rows(vLandmarksInner));
				mat Vp, Sp, Up;
				double dTolerance    = 1e-6;
		        unsigned int rank;

				{
/*					pinv_ny(Cnm, Wpinv, vLandmarksInner, dTolerance, Vp, Sp, Up, rank);


					mat  vTemp1Scaled    =  (Z.t()*K1FullIndex);
    				mat  vTemp2Scaled    =  (I.t()*K1FullIndex);
	
					vec  wPart2			 = CurrentTrainY % (vTemp1Scaled * K2Nz).t() - b * CurrentTrainY % (vTemp2Scaled * K2Nz).t();
			        Temp				 = (vT % wPart2);
    				Temp2				 = Sp * Up.cols(0,rank-1).t();
    				Vp				     = Vp.cols(0,rank-1);
					w(nonZero)			 = (Vp*(Temp2*Temp)) % vT;
*/
				}
				{
					mat  vTemp1Scaled    = (Z.t()*K1FullIndex);
    				mat  vTemp2Scaled    = (I.t()*K1FullIndex);
					mat  K2Nz            = K2.cols(nonZero);
					vec  wPart2			 = CurrentTrainY % (vTemp1Scaled * K2Nz).t() - b * CurrentTrainY % (vTemp2Scaled * K2Nz).t();
					K2Nz.reset();
					vTemp1Scaled.reset();
					vTemp2Scaled.reset();					
		        	Temp				 = (vT % wPart2);

					
					pinv_ny_sym(Cnm, Wpinv, vLandmarksInner, dTolerance, Sp, Up, rank);
					Cnm.reset();
					Wpinv.reset();
					
    				mat Temp2			 = Sp * Up.cols(0,rank-1).t();
					Sp.reset();
					Up				     = Up.cols(0,rank-1);
													
					w(nonZero)			 = (Up*(Temp2*Temp)) % vT;

				}

			}

		    vec vTemp     = CurrentTrainY % w(nonZero);
			mat Temp	  = vTemp.t() * K1.rows(nonZero); 
			vec bPart2    = (Temp * K2FullIndex).t();
			Temp.reset();
			b = S*pow(1+ S*ndata*S,-1)*S*as_scalar(I.t()*Z - I.t()*bPart2); // % Eq 52                             

			nonZero	      = find(w > w_minimal);			
			if(nonZero.n_elem < 1)
			{
				// reset w and provide a dummy model 
				// can happen for ill posed kernel approximations
				w       = randu(ndata,1) + 0.2;
				nonZero	= find(w > w_minimal);
				n       = niters; // cancel loop
			}

			// determine used vectors
			uvec used     = nonZero;

			vec wNew      = zeros(ndata,1);
			wNew(nonZero) = w(nonZero);
			w             = wNew;
		
			// non-zero weight vector
			w_nz          = w(nonZero);	

			if (n >1 && max(abs(w - wold))< threshold)
    			break;
			else
    			wold = w; 
		
	}
	m_oPCVMModel = pcvm_model(w % Label,b,oKernel.GetKernelType()); 
}

pcvm_parameters pcvm_train_ny::GetParameters(void)
{
	return m_oParameters;
}
