/**
 * @file pcvm_test.cpp
 * @author Frank-Michael Schleif
 *
 * Test the pcvm classifier on the iris data to test
 * expected model parameters against obtained ones.
 *
 */
#include "../mlstub/load.hpp"
#include <boost/bind.hpp>
#include <boost/test/unit_test.hpp>
#include "old_boost_test_definitions.hpp"
#include "pcvmtrain.h"
#include "kernel-matrix.h"
#include "pcvmmulticlassmodel.h"
#include "math_stat_helper.hpp"
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>



BOOST_AUTO_TEST_SUITE(PCVMMethodTest);


/**
 * Make sure that the model obtained on the iris data fits to the expected reference model
 */
BOOST_AUTO_TEST_CASE(PCVMIrisTest)
{
  // Load the dataset.
  arma::mat DataT;
  mlpack_stub::data::Load("./data/iris_kernel.csv", DataT, true,false);

  int M = DataT.n_rows;
  int N = DataT.n_cols;
	
  arma::vec Label = vectorise(DataT.col(N-1));  
  DataT = DataT.cols(0,N-2);
  arma::vec vUniqueLabel = unique(Label);
  
  pcvm_parameters oParameters;
  oParameters.SetMinimalThreshold(1e-3);
  oParameters.SetConvergenceThreshold(1e-3);
  oParameters.SetIterations (500);
	
  pcvm_train oPCVMTrain;
  oPCVMTrain.SetParameters (oParameters );
  KernelMatrix oKernel(2);
	
  bool bResult = false;
  if(oKernel.GetKernelType () != 2)
		bResult = oKernel.Initialize(DataT,DataT);
  else
  {
		bResult = oKernel.Initialize(DataT,DataT,DataT);
  }  		
  assert(bResult); 

  // run directly the multiclass 1 vs rest wrapper	
  pcvm_multiclass_model sFullModel;
  sFullModel.GenerateMultiClassModels(oPCVMTrain, oKernel, DataT, Label);

  // Create an output archive
  std::ofstream ofs("test.model");
  boost::archive::text_oarchive ar(ofs);

  // Write data
  ar & sFullModel;

	
  // predict the outcomes for the current trained full model
  mat TrainingData = DataT.rows(sFullModel.sm_vReferenceIndices);	  
  std::cout << "* Used model indices:" << sFullModel.sm_vReferenceIndices.t() << std::endl;
  if(oKernel.GetKernelType () != 2)
		bResult = oKernel.Initialize(TrainingData,DataT);
  else
  {
		bResult = oKernel.Initialize(TrainingData,DataT,DataT);
  }
  TrainingData.reset();
  assert(bResult); 
  mat PredictionResults = sFullModel.CalculateModelPredictions(oKernel);  
  uvec vPredictedModel  = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
	  
  double dMeanError =  (double)sum(abs(sFullModel.sm_vUniqueLabel(vPredictedModel)-Label)<std::numeric_limits<double>::epsilon())/(double)Label.size();
  std::cout << dMeanError << std::endl;
  BOOST_CHECK( dMeanError >=0.9 );
}

BOOST_AUTO_TEST_SUITE_END();
